from numpy import arcsin, around, power, sqrt


def amplitude(state, circuit_size, show_bit=-1, round: int = 3, radian: bool = False):
    if round < 0:
        exit(f"Error: QuantumCircuit.tools.amplitude() -- round placement must be a value greater than 0.")
    if type(show_bit) == int and show_bit < 0:
        # for loop will go through this statement 2^n times for _state length
        # converts state to: (x^2 + y^2) where x is real values and y is imaginary values
        statevector = sqrt(power(state.real, 2) + power(state.imag, 2))

    # converts unsigned bit representation into a integer if string.
    # handles error for being out of bounds of 2^(num_of_qubits)
    # converts string bit to an integer to find the value for
    elif type(show_bit) == str or type(show_bit) == int:
        statevector = show_bit
        if type(show_bit) == str:
            statevector = int(show_bit, 2)

        if 2**circuit_size <= statevector:
            exit(
                f"Error: QuantumCircuit.tools.amplitude() -- Called bit to find amplitude is not within range of possible values."
            )
        statevector = sqrt(power(state[statevector].real, 2) + power(state[statevector].imag, 2))
    # wrong value inputted for finding the show_bit
    else:
        exit(
            f"Error: QuantumCircuit.tools.amplitude() -- show_bit given wrong type of value, binary string or integer are only allowed."
        )
    if radian:
        statevector = arcsin(statevector) * 2
    return around(statevector, decimals=round)


from qsm.qubit import *

import scipy.sparse as ss

__all__ = [
    "gen_qubit",

]

class QuantumCircuit:
    def __init__(self, qbits, state: str="z"):

        self.circuit_size = qbits

        self.state = ss.csr_matrix(gen_qubit(state))

        for i in range(qbits-1):
            self.state = ss.kron(self.state, ss.csr_matrix(gen_qubit(state)))

    def apply_gate(self, gate, target_qubit: int):
        gate_class = gate([self.state], self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def apply_controlled_gate(self, gate, target_qubit, control_qubit):
        gate_class = gate([self.state], self.circuit_size)
        self.state = gate_class.apply(target_qubit, control_qubit)

    def custom_controlled_gate(self, gate_matrix, target_qubit, control_qubit):
        gate_class = ControlledGate([self.state], gate_matrix, self.circuit_size)
        self.state = gate_class.apply(target_qubit, control_qubit)

    def u(self, target_qubit, theta, phi, lamba):
        gate_class = U([self.state], theta, phi, lamba, self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def p(self, target_qubit, theta):
        gate_class = P([self.state], theta, self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def rz(self, target_qubit, theta):
        gate_class = RZ([self.state], theta, self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def rx(self, target_qubit, theta):
        gate_class = RX([self.state], theta, self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def ry(self, target_qubit, theta):
        gate_class = RY([self.state], theta, self.circuit_size)
        self.state = gate_class.apply(target_qubit)

    def get_state(self):
        return self.state.toarray()

    def prob(self, show_percent: bool = False):
        return probabilities(self.state.toarray(), self.circuit_size, show_percent)

    def measure(self):
        prob = self.prob()
        return measure(self.state.toarray(), self.circuit_size, prob)

    def phase_angle(self):
        return phaseangle(self.state.toarray(), self.circuit_size)

    def amplitude(self):
        return amplitude(self.state.toarray(), self.circuit_size)



from numpy import abs, around, multiply, square


def probabilities(state, circuit_size, show_percent: bool, show_bit: int = -1, round: int = 3):
    # determines if round is less than 0
    # if the
    if round < 0:
        exit(f"Error: QuantumCircuit.tools.probabilities() -- round placement must be a value greater than 0.")

    if type(show_bit) == int and show_bit < 0:
        probability = square(abs(state.flatten()))
    elif type(show_bit) == str or isinstance(show_bit, int):
        if type(show_bit) == str:
            show_bit = int(show_bit, 2)

        if 2**circuit_size <= show_bit:
            exit(
                f"Error: QuantumCircuit.tools.probabilities() -- Called bit to find phase angle is not within range of possible values."
            )

        probability = square(abs(state.flatten()[show_bit]))

    else:
        exit(
            f"Error: QuantumCircuit.tools.probabilities() -- show_bit given wrong type of value, binary string or integers in range are only allowed."
        )

    # get the probabilties for the "winner" of the measurement at any
    # single point on the circuit
    # collapse the circuit into 1D
    # square the values to get the probabilities of each qubit state
    # turn all the complex numbers into real values
    if show_percent:
        probability = multiply(probability, 100)
    return around(probability, decimals=round)


import numpy as np
from scipy import sparse
from qsm.method import *


class Gate:
    def __init__(self, qbits, apply_matrice, circuit_size):
        self.qbits = qbits
        self.gate = apply_matrice
        self.circuit_size = circuit_size

    def __operator_matrix__(self, gate_matrix: np.array, qubit: int):
        i_matrix = identity()
        gate_queue = [i_matrix for _ in range(self.circuit_size)]
        gate_queue[qubit] = gate_matrix

        operator_matrix = sparse.csr_matrix(gate_queue[0])

        for gate in gate_queue[1:]:
            operator_matrix = sparse.kron(operator_matrix, sparse.csr_matrix(gate))

        return operator_matrix

    def apply(self, qbit):
        new_array = self.__operator_matrix__(self.gate, qbit).dot(self.qbits[0])
        self.qbits[0] = new_array
        return new_array


class ControlledGate:
    def __init__(self, qbits, apply_matrice, circuit_size):
        self.qbits = qbits
        self.circuit_size = circuit_size
        self.gate = apply_matrice

    def __controlled_phase_handler__(self, matrix_to_calculate: np.array, control_qubit: int, target_qubit: int):
        if matrix_to_calculate.shape != (2, 2):
            matrix_to_calculate = matrix_to_calculate[-2:, -2:]

        bra_ket_zero = np.array([[1 + 0j, 0 + 0j], [0 + 0j, 0 + 0j]], "F")

        bra_ket_one = np.array([[0 + 0j, 0 + 0j], [0 + 0j, 1 + 0j]], "F")

        bra_ket_zero_kron = [identity()] * self.circuit_size

        bra_ket_one_kron = [identity()] * self.circuit_size

        bra_ket_zero_kron[control_qubit] = bra_ket_zero
        bra_ket_one_kron[control_qubit] = bra_ket_one
        bra_ket_one_kron[target_qubit] = matrix_to_calculate

        to_add_zero = sparse.csr_matrix(bra_ket_zero_kron[0])

        to_add_one = sparse.csr_matrix(bra_ket_one_kron[0])

        for i in range(1, len(bra_ket_zero_kron)):
            to_add_zero = sparse.kron(to_add_zero, sparse.csr_matrix(bra_ket_zero_kron[i]))

            to_add_one = sparse.kron(to_add_one, sparse.csr_matrix(bra_ket_one_kron[i]))

        to_add_zero += to_add_one
        matrix = to_add_zero.dot(self.qbits[0])
        return matrix

    def apply(self, qbit_target, qbit_control):
        new_array = sparse.csr_matrix(self.__controlled_phase_handler__(self.gate, qbit_control, qbit_target))
        return new_array


CX = lambda qbits, circuit_size: ControlledGate(qbits, paulix(), circuit_size)
I = lambda qbits, circuit_size: Gate(qbits, identity(), circuit_size)
X = lambda qbits, circuit_size: Gate(qbits, paulix(), circuit_size)
Y = lambda qbits, circuit_size: Gate(qbits, pauliy(), circuit_size)
Z = lambda qbits, circuit_size: Gate(qbits, pauliz(), circuit_size)
H = lambda qbits, circuit_size: Gate(qbits, hadamard(), circuit_size)
P = lambda qbits, theta, circuit_size: Gate(qbits, phase(theta), circuit_size)
S = lambda qbits, circuit_size: Gate(qbits, s(), circuit_size)
Sdg = lambda qbits, circuit_size: Gate(qbits, sdg(), circuit_size)
T = lambda qbits, circuit_size: Gate(qbits, t(), circuit_size)
Tdg = lambda qbits, circuit_size: Gate(qbits, tdg(), circuit_size)
RZ = lambda qbits, theta, circuit_size: Gate(qbits, rz(theta), circuit_size)
RX = lambda qbits, theta, circuit_size: Gate(qbits, rx(theta), circuit_size)
RY = lambda qbits, theta, circuit_size: Gate(qbits, ry(theta), circuit_size)
SX = lambda qbits, circuit_size: Gate(qbits, sx(), circuit_size)
SXdg = lambda qbits, circuit_size: Gate(qbits, sxdg(), circuit_size)
U = lambda qbits, theta, phi, lamba, circuit_size: Gate(qbits, u(theta, phi, lamba), circuit_size)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='qsm',
    version='0.1.0',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy'
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import numpy as np

def qubit(initial_state: chr = "z"):
    """
    Args:
        initial_state:
            A character input of either x, y, z to indicate either the ]
            qubit's facing to be at either the z axis, y axis, or x axis in
            true whole value.
    """
    # Sets qubit to the z-axis and will face "upwards" or [1,0,...n^2 - 1].
    if initial_state == "z":
        return np.array([[1 + 0j], [0 + 0j]], "F")
        # Sets the qubit to the y-axis.
    elif initial_state == "y":
        return np.array([[0 + 0j + 1 + 0j], [(1 + 0j + 0 + 0j) * 0 + 1j]], "F") / np.sqrt(2)
        # Sets the qubit to the y-axis.
    elif initial_state == "x":
        return np.array([[0 + 0j + 1 + 0j], [1 + 0j + 0 + 0j]], "F") / np.sqrt(2)
        # Calls error as no other plane of existence for calculations exists.
    else:
        exit(
            f"Error: Qubit.__init__() -- Qubit must be initialized in a 0 \
                or 1 state \n Qubit was initialized in state: {initial_state}."
        )


from numpy import arange, random


def measure(state, circuit_size, probabilities):
    # randomly selects the measured state using self.probabilities()
    # number of bits
    # creates a list of bits in strings
    # numpy.random.choice takes in the list we will select from, size of the returning list,
    #  and p = weights of each element
    final_state = random.choice(arange(len(state), dtype=int), 1, p=probabilities)
    # take out the bits from the returned list and convert to a string
    final_state = str(bin(final_state[0]))[2:]
    # pad with zeroes if needed
    final_state = final_state.zfill(circuit_size)
    # return the final state
    return final_state


from numpy import angle, around, mod, pi


def phaseangle(state, circuit_size, show_bit=-1, round: int = 3, radian: bool = True):
    # if rounding value is less than 0, exits as this is improper.
    if round < 0:
        exit(f"Error: QuantumCircuit.tools.phaseAngle() -- round placement must be a value greater than 0.")

    if type(show_bit) == int and show_bit < 0:
        phaseangle = mod(angle(state), 2 * pi) * (180 / pi)

    elif type(show_bit) == str or type(show_bit) == int:
        phaseangle = show_bit

        if type(show_bit) == str:
            phaseangle = int(show_bit, 2)
        if 2**circuit_size <= phaseangle:
            exit(
                f"Error: QuantumCircuit.tools.phaseAngle() -- Called bit to find phase angle is not within range of possible values."
            )

        phaseangle = mod(angle(state[phaseangle]), 2 * pi) * (180 / pi)
    else:
        exit(
            f"Error: QuantumCircuit.tools.phaseAngle() -- show_bit given wrong type of value, binary string or integers in range are only allowed."
        )

    if radian:
        phaseangle *= pi / 180

    return around(phaseangle, decimals=round)